class HttpException(Exception):
    pass

class UndefinedToken(Exception):
    pass

class RateLimited(HttpException):
    def __init__(self, json, headers):
        self.json = json
        self.headers = headers
        self.message = self.json["message"]
        self.retry_after = self.json["retry_after"]
        super().__init__(self.message)

class Unauthorized(HttpException):
    pass

class AccessDenied(HttpException):
    pass



from requests_oauthlib import OAuth2Session
from flask import request
from .exceptions import *
import requests

discord_api_base = "https://discordapp.com/api/v9"
discord_api_authorize = discord_api_base + "/oauth2/authorize"
discord_api_token = discord_api_base + "/oauth2/token"
discord_api_users = discord_api_base + "/users/@me"

class oauth:
    def __init__(self, flaskapp, db, DiscordOauth2):
        self.client_id = flaskapp.flask_keys["FLASKORD_CLIENT_ID"]
        self.client_secret = flaskapp.flask_keys["FLASKORD_CLIENT_SECRET"]
        self.redirect_uri = flaskapp.flask_keys["FLASKORD_OAUTH_REDIRECT_URI"]
        self.scope = flaskapp.flask_keys["FLASKORD_OAUTH_SCOPE"]
        self.db = db
        self.DiscordOauth2 = DiscordOauth2

    def get_login_url(self) -> str:
        oauth = OAuth2Session(self.client_id, redirect_uri=self.redirect_uri, scope=self.scope)
        login_url, state = oauth.authorization_url(discord_api_authorize)
        return login_url

    def confirm_login(self) -> dict:
        payload = {
           "client_id": self.client_id,
           "client_secret": self.client_secret,
           "grant_type": "authorization_code",
           "code": request.args.get("code"),
           "redirect_uri": self.redirect_uri,
           "scope": self.scope
        }

        try:
            access_token = requests.post(url = discord_api_token, data = payload).json()["access_token"]
            headers = {"Authorization": f"Bearer {access_token}"}

            user = requests.get(url = discord_api_users, headers = headers).json()

            if not self.DiscordOauth2.query.filter_by(discord_id=user["id"]).first(): # check if user already registered
                self.db.session.add(self.DiscordOauth2(discord_id=user["id"], discord_name=f'{user["username"]}#{user["discriminator"]}'))
                self.db.session.commit()

            return user
        except KeyError:
            return None



from .flaskord import client
from .exceptions import UndefinedToken
from .oauth2 import oauth
from flask_sqlalchemy import SQLAlchemy


base_keys = sorted(['FLASKORD_CLIENT_ID', 'FLASKORD_CLIENT_SECRET', 'FLASKORD_TOKEN'])
oauth_keys = sorted(['FLASKORD_OAUTH_REDIRECT_URI', 'FLASKORD_OAUTH_SCOPE', 'SQLALCHEMY_DATABASE_URI'])

class Flaskord(object):
    def __init__(self, flaskapp = None):
        if flaskapp is not None:
            self.flask_keys = flaskapp.config

            if sorted(list(set(list(base_keys)) & set(list(flaskapp.config.keys())))) != base_keys:
                raise KeyError(f"Some flask app keys not found in project. Make sure you have set up app['FLASKORD_CLIENT_ID'], app['FLASKORD_CLIENT_SECRET'] and app['FLASKORD_TOKEN']")

            self.client = client(flaskapp.config["FLASKORD_TOKEN"])

            if "FLASKORD_OAUTH" in list(flaskapp.config.keys()):
                if sorted(list(set(list(oauth_keys)) & set(list(flaskapp.config.keys())))) != oauth_keys:
                    raise KeyError(f"Some flask app keys not found in project. Make sure you have set up app['FLASKORD_OAUTH_REDIRECT_URI'], app['FLASKORD_OAUTH_SCOPE'] and app['SQLALCHEMY_DATABASE_URI']")

                db = SQLAlchemy(flaskapp)

                class DiscordOauth2(db.Model):
                    id = db.Column(db.Integer, primary_key=True)
                    discord_id = db.Column(db.Integer, nullable=False)
                    discord_name = db.Column(db.String, nullable=False)

                    def __repr__(self):
                        return '<DiscordOauth2 %r>' % self.id

                self.oauth = oauth(self, db, DiscordOauth2)

                db.create_all()

        else:
            raise UndefinedToken("app is not set")



import requests
import json
from .exceptions import *
from typing import Union

# Read https://discord.com/developers/docs/resources/guild

discord_api_base = "https://discordapp.com/api/v9"
discord_api_authorize = discord_api_base + "/oauth2/authorize"
discord_api_token = discord_api_base + "/oauth2/token"

class client:
    def __init__(self, token):
        self.headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            "authorization": f"Bot {token}",
            "intents": "GUILD_MEMBERS" # it doesn't need it, but I'll leave it here :/
        }

    def __run__(self, route, method, payload=None):
        route = discord_api_base + route
        resp = requests.request(method, route, headers=self.headers) if payload is None else requests.request(method, route, data=json.dumps(payload), headers=self.headers)

        if resp.status_code == 401:
            raise Unauthorized("Token incorrect")

        elif resp.status_code == 429:
            raise RateLimited(resp.json(), resp.headers)

        else:
            return resp.json()


    # base methods

    def get_user(self, # get basic user information
                 user_id: Union[str, int]) -> dict:

        resp =  self.__run__(route=f"/users/{user_id}", method="GET")
        avatar = f"https://cdn.discordapp.com/avatars/{user_id}/{resp['avatar']}.png?size=512"
        resp["avatar_url"] = avatar
        return resp


    def fetch_member(self, # fetching user on guild
                     guild_id: Union[str, int],
                     user_id: Union[str, int]) -> dict:

        resp = self.__run__(route=f"/guilds/{guild_id}/members/{user_id}", method="GET")
        avatar = f"https://cdn.discordapp.com/avatars/{user_id}/{resp['avatar']}.png?size=512"
        resp["avatar_url"] = avatar
        return resp


    def get_guild(self, # get basic information about guild
                  guild_id: Union[str, int]) -> dict:

        resp = self.__run__(route=f"/guilds/{guild_id}?with_counts=true", method='GET')
        resp["member_count"] = resp["approximate_member_count"]
        return resp

    # advanced methods

    def send(self, payload: Union[dict, str], # sends message in user DM or in channel
             channel_id: Union[str, int] = None,
             user_id: Union[str, int] = None):

        if channel_id or user_id:
            channel_id = self.__run__(route="/users/@me/channels", method="POST", payload={"recipients": [int(user_id)]})["id"] if user_id else channel_id

            payload = {"content": payload} if not isinstance(payload, dict) else payload
            return self.__run__(route=f"/channels/{channel_id}/messages", method="POST", payload=payload)
        else:
            return HttpException("user_id or channel_id must be set")


    def add_roles(self, # add roles for user on guild (easier method than edit_member)
                  guild_id: Union[str, int],
                  user_id: Union[str, int],
                  role_id: Union[str, int]):

        role_id = str(role_id)
        try:
            roles = self.fetch_member(guild_id, user_id)["roles"]

            if role_id not in roles: roles.append(role_id)

            payload = {"roles": roles}

            return self.__run__(route=f"/guilds/{guild_id}/members/{user_id}", method="PATCH", payload=payload)
        except KeyError:
            return HttpException("Member not found")


    def remove_roles(self, # remove roles for user on guild (easier method than edit_member)
                     guild_id: Union[str, int],
                     user_id: Union[str, int],
                     role_id: Union[str, int]):

        role_id = str(role_id)
        try:
            roles = self.fetch_member(guild_id, user_id)["roles"]
            if role_id in roles: roles.remove(role_id)
            payload = {"roles": roles}

            return self.__run__(route=f"/guilds/{guild_id}/members/{user_id}", method="PATCH", payload=payload)

        except KeyError:
            return HttpException("Member not found")


    def create_channel(self,
                       guild_id: Union[str, int],
                       channel: dict):

        return self.__run__(route=f"/guilds/{guild_id}/channels", method="POST", payload=channel)


    def delete_channel(self,
                    guild_id: Union[str, int],
                    channel_id: Union[str, int]):

        return self.__run__(route=f'/guilds/{guild_id}/channels/{channel_id}', method='DELETE')


    def edit_channel(self,
                     guild_id: Union[str, int],
                     channel_id: Union[str, int],
                     payload: dict):

        return self.__run__(route = f"/channels/{channel_id}", method="PATCH", payload=payload)


    def create_role(self,
                    guild_id: Union[str, int],
                    role: dict):

        return self.__run__(route=f'/guilds/{guild_id}/roles', method="POST", payload=role)


    def delete_role(self,
                    guild_id: Union[str, int],
                    role_id: Union[str, int]):

        return self.__run__(route=f'/guilds/{guild_id}/roles/{role_id}', method='DELETE')


    def edit_role(self,
                  guild_id: Union[str, int],
                  role_id: Union[str, int],
                  payload: dict):

        return self.__run__(route=f'/guilds/{guild_id}/roles/{role_id}', method="PATCH", payload=payload)


    def kick_member(self,
                    guild_id: Union[str, int],
                    user_id: Union[str, int]):

        return self.__run__(route=f'/guilds{guild_id}/members/{user_id}', method="DELETE")


    def ban_member(self,
                   guild_id: Union[str, int],
                   user_id: Union[str, int]):

        return self.__run__(route=f'/guilds/{guild_id}/bans/{user_id}', method='PUT')


    def unban_member(self,
                   guild_id: Union[str, int],
                   user_id: Union[str, int]):

        return self.__run__(route=f'/guilds/{guild_id}/bans/{user_id}', method='DELETE')


    def edit_member(self,
                    guild_id: Union[str, int],
                    user_id: Union[str, int],
                    payload: dict):

        return self.__run__(route=f"/guilds/{guild_id}/members/{user_id}", method="PATCH", payload=payload)
